/* ftindex.c
   Given a set of files containing family tree information,
   generate a sorted index file "index.ft" suitable for 
   printing with "geneal"

   Author: Graham Freeman
   Copyright: Graham Freeman, March 1994.
   This program may not be used for commercial purposes without
   the consent of the author.  It may be freely transmitted 
   provided this authorship and copyright notice is retained 
   unmodified.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define Bsize 1025
enum Pstate {FindAnc, FirstAnc, OtherAnc, BuildAnc, SkipAnc, GenNames};
struct varray
{
   int  maxnlines;
   int  count;
   char **lines;
};

char *getfirstname(char *buf);
void indexentry(char *name, char *anctext, struct varray *ind);
void surnamefirst(char *name);
void removewhiteend(char *buf);
int safecat(char *str1, register char *str2, int size1);
void outputfileheader(FILE *ind);
void mergedat(FILE *fin, FILE *fout, char *line[], int nline);
int compar(const void* first, const void* second);
int getline(register char *line, int nmax, register FILE *fp);

/***************************************************************/
int main(int argc, char *argv[])
{
   FILE           *ft, *ind, *f[2];
   char           *fname[2] = {"scrtmpf1.ftf",
                               "scrtmpf2.ftf"};
   int            nerrs = 0;
   int            file;
   int            currentfile = -1;
   int            nred;
   int            i;
   char           buf[Bsize];
   enum Pstate    state;
   char           *match, *name, *nonbl;
   char           anctext[2048];
   struct varray  indx = {900,0};

               /* initial checking */
   if (argc < 2)
   {
      fprintf(stderr, "ftindex must have arguments, the files "
                      "containing family tree data for building the index\n");
      exit(1);
   }
   if ((ind = fopen("index.qs","w")) == NULL)
   {
      fprintf(stderr, "ftindex cannot open the index file \"index\"\n");
      exit(1);
   }
   for (file=1; file < argc; ++file)
   {
      if ((ft = fopen(argv[file],"r")) == NULL)
      {
         fprintf(stderr, "ftindex cannot open the tree file %s\n",
            argv[file]);
         ++nerrs;
      }
      else
         fclose(ft);
   }
   if ((f[0] = fopen(fname[0],"w")) == NULL)
   {
      fprintf(stderr, "ftindex cannot create temporary file %s\n",
         fname[0]);
      ++nerrs;
   }
   if ((f[1] = fopen(fname[1],"w")) == NULL)
   {
      fprintf(stderr, "ftindex cannot create temporary file %s\n",
         fname[1]);
      ++nerrs;
   }
   else fclose(f[1]);
   if ((indx.lines = (char**)malloc(indx.maxnlines*sizeof(char*))) == NULL)
   {
      fprintf(stderr,"ftindex: unable to allocate memory for sorting index\n");
      ++ nerrs;
   }
   if (nerrs > 0) exit(1);

               /* main loop, processing each file, sorting and merging into
                  a single index file */
   for (file=1; file < argc; ++file)
   {
      ft = fopen(argv[file],"r");
      state = FindAnc;
      while ((nred = getline(buf,Bsize,ft)) > -1)
      {
         switch (state)
         {
            case FindAnc:     /* skip all looking for "<ancestry>" */
               if (strncmp(buf,"<ancestry>",10) != 0) continue;
               state = FirstAnc;
               break;
            case FirstAnc:    /* initialize the ancestry string */
               name = getfirstname(buf);
               strcpy(anctext,name);
               state = BuildAnc;
               break;
            case OtherAnc:    /* add another ancestry to existing ancestry 
                                 string */
               name = getfirstname(buf);
               safecat(anctext,name,sizeof(anctext));
               state = BuildAnc;
               break;
            case BuildAnc:    /* build a text string of the ancestry */
               i = strspn(buf," \t");
               nonbl = &buf[i];
               if (strncmp(nonbl,"<>",2) == 0)
               {
                  if (strpbrk(nonbl,"=~") == NULL) continue;
                  name = getfirstname(&nonbl[2]);
                  safecat(anctext," / ",sizeof(anctext));
                  safecat(anctext,name,sizeof(anctext));
               }
               else if (strncmp(nonbl,"<*>",3) == 0)
               {
                  name = getfirstname(&nonbl[3]);
                  safecat(anctext," / ",sizeof(anctext));
                  safecat(anctext,name,sizeof(anctext));
               }
               else if (strncmp(nonbl,"<generation>",12) == 0)
               {
                  state = GenNames;
                  continue;
               }
               else if (strncmp(nonbl,"<ancestry>",10) == 0)
               {
                  /* state = SkipAnc; *//*Only use the first ancestry */
                  /* continue; */
                  safecat(anctext,"#>",sizeof(anctext)); /*Store each ancestry*/ 
                  state = OtherAnc;
                  continue;
               }
               else
               {
                  continue;
               }
               break;
            case SkipAnc:     /* skip any ancestry other than the first */
               i = strspn(buf," \t");
               nonbl = &buf[i];
               if (strncmp(nonbl,"<generation>",12) == 0)
               {
                  state = GenNames;
                  continue;
               }
               break;
            case GenNames:    /* output names for the index */
               i = strspn(buf," \t");
               nonbl = &buf[i];
               if (strncmp(nonbl,"<>",2) == 0 || strncmp(nonbl,"<*>",3) == 0)
               {
                  name = getfirstname(&nonbl[2]);
                  surnamefirst(name);
                  if (isalpha(name[0]))
                     indexentry(name,anctext,&indx);
               }
               else if (strncmp(nonbl,"<comment>",9) == 0)
               {
                  state = FindAnc;
                  continue;
               }
               else if (strncmp(nonbl,"<ancestry>",10) == 0)
               {
                  state = FirstAnc;
                  continue;
               }
               else
               {
                  continue;
               }
               break;
         }
      }
      fclose(ft);

      qsort(indx.lines,indx.count,sizeof(char*),compar);

      if (currentfile == -1)
      {
         for (i=0; i<indx.count; ++i)
         {
            fputs(indx.lines[i],f[0]);
            fputc('\n',f[0]);
         }
         fclose(f[0]);
         f[0] = fopen(fname[0],"r");
         f[1] = fopen(fname[1],"w");
      }
      else
      {
         mergedat(f[currentfile], f[1-currentfile], indx.lines, indx.count);
         fclose(f[currentfile]);
         fclose(f[1-currentfile]);
         f[currentfile] = fopen(fname[currentfile],"w");
         f[1-currentfile] = fopen(fname[1-currentfile],"r");
      }
      currentfile = (currentfile + 1)%2;
      for (i=0; i<indx.count; ++i)
      {
         free(indx.lines[i]);
      }
      indx.count = 0;
   }

         /* now copy the sorted composite file to the output file, 
            adding suitable layout instructions */
   fprintf(ind,"%%BD%%%%CL%%Index\n");
   fprintf(ind,"%%LT%%%%SZ,9%%%%P,25,-25,-0.7%%%%P%%%%NC,2%%\n");
   while (getline(buf,sizeof(buf),f[currentfile]) > 0)
   {
      char *pt = strchr(buf,'#'), 
           *bf = buf;
      *pt = '\0';
      fprintf(ind,"%s     %%TR,25%%",bf);
      bf = pt+1;
      while ((pt = strchr(bf,'#')) != NULL)
      {
         *pt = '\0';
         fprintf(ind,"%s%%L%%%%TR,25%%",bf);
         bf = pt+1;
      }
      fprintf(ind,"%s\n%%P%%",bf);
   }
   fprintf(ind,"\n");
   fclose(f[0]);
   fclose(f[1]);
   remove(fname[0]);
   remove(fname[1]);
   fclose(ind);
}

/***************************************************************/
void surnamefirst(char *name)
   /* given a name with either the surname last, or 
      the surname second-last followed by a comment in parentheses,
      reverse the order, so that the surname comes first, followed
      by a comma, and the remainder of the name. "name" will be
      overwritten in the process. 
      If the last name looks like "*", this should be removed and
      the previous name used instead. */
{
   char temp[Bsize];
   char *pt, *pt2, *pt3=NULL;
   int  pos;
   pos = strlen(name)-1;
   if (name[pos] == ')')
   {
      pt3 = strrchr(name,'(');
      if (pt3 == name) return;
      if (pt3 > name) *pt3 = '\0';
      removewhiteend(name);
   }
   pt = strrchr(name,' ');
   if (pt == NULL) return;
   *pt = '\0';
   removewhiteend(name);
   if (*(pt+1) == '*' || *(pt+1) == '?')
   {
      pt = strrchr(name,' ');
      if (pt == NULL) return;
      *pt = '\0';
      removewhiteend(name);
   }
   if (*(pt+1) == '(')
   {
      pt2 = pt;
      pt = strrchr(name,' ');
      if (pt != NULL) 
      {
         *pt = '\0';
         strcpy(temp,pt+1);
         strcat(temp,", ");
      }
      strcat(temp,name);
      if (*(pt2+1) == '(')
      {
         strcat(temp," ");
         strcat(temp,pt2+1);
      }
   }
   else
   {
      *pt = '\0';
      strcpy(temp,pt+1);
      strcat(temp,", ");
      strcat(temp,name);
   }
   if (pt3 != NULL)
   {
      strcat(temp," (");
      strcat(temp,pt3+1);
   }
   strcpy(name,temp);
}

/***************************************************************/
void indexentry(char *name, char *anctext, struct varray *ind)
   /* output the name followed by the anctext, separated by a tab 
      character, to the file ind */
{
               /* allocate a larger buffer if we will spill over */
   if (ind->count >= ind->maxnlines)
   {
      char **temp;
      int  i;
      ind->maxnlines *= 2;
      temp = (char**)malloc(ind->maxnlines*sizeof(char*));
      for (i=0; i<ind->count; ++i)
         temp[i] = ind->lines[i];
      free(ind->lines);
      ind->lines = temp;
   }
               /* place the text in the next slot */
   ind->lines[ind->count] = (char *) malloc(strlen(name) + strlen(anctext) + 2);
   strcpy(ind->lines[ind->count],name);
   strcat(ind->lines[ind->count],"#");
   strcat(ind->lines[ind->count],anctext);
   ++ ind->count;
}

/***************************************************************/
void mergedat(FILE *fin, FILE *fout, char *line[], int nline)
{
   char buf[Bsize];
   int  end1=0, end2=0, cmp, count=0;
   getline(buf,sizeof(buf),fin);
   while (!end1 && !end2)
   {
      cmp = strcmp(buf,line[count]);
      if (cmp < 0)
      {
         fputs(buf,fout);
         fputc('\n',fout);
         if (getline(buf,sizeof(buf),fin) <= 0) end1 = 1;
      }
      else if (cmp > 0)
      {
         fputs(line[count++],fout);
         fputc('\n',fout);
         if (count >= nline) end2 = 1;
      }
      else  /* == */
      {
         fputs(buf,fout);
         fputc('\n',fout);
         if (getline(buf,sizeof(buf),fin) <= 0) end1 = 1;
         if (count >= nline) end2 = 1;
      }
   }
   while (!end1)
   {
      fputs(buf,fout);
      fputc('\n',fout);
      if (getline(buf,sizeof(buf),fin) <= 0) end1 = 1;
   }
   while (!end2)
   {
      fputs(line[count++],fout);
      fputc('\n',fout);
      if (count >= nline) end2 = 1;
   }
}

/***************************************************************/
int compar(const void* first, const void* second)
{
   char **a = (char **)first, **b = (char**)second;
   return(strcmp(*a, *b));
}

/***************************************************************/
void outputfileheader( FILE *ind )
{
   fprintf(ind,"%%BD%%%%CL%%Index\n");
   fprintf(ind,"%%LT%%%%SZ,9%%%%P,25,-25,-0.7%%%%P%%%%NC,2%%\n");
}

/***************************************************************/
char *getfirstname(char *buf)
   /* Extract the first string from the line before any '=', '~' or '['.
      Remove leading and trailing whitespace.
      Return the address within 'buf' of the string.
      'buf' will be damaged in the process.
   */
{
   int end, start, i;
   end = strcspn(buf,"=~[");
   buf[end] = '\0';
   removewhiteend(buf);
   start = strspn(buf," \t");
   return(&buf[start]);
}

/***************************************************************/
void removewhiteend(char *buf)
   /* replace any trailing white space with '\0' */
{
   int end;
   for (end = strlen(buf)-1; end > 0; --end)
      if (buf[end] == ' ' || buf[end] == '\t') buf[end] = '\0';
      else break;
}

/***************************************************************/
int safecat(char *str1, register char *str2, int size1)
   /* strcat, but check that the length of the destination does
      not exceed 'size1' (including '\0').
      Return 1 if successful, else zero.
   */
{
   register int i;
   for (i=0; i<size1; ++i)
      if (str1[i] == '\0') break;
   while (i<size1 && *str2 != '\0')
   {
      str1[i++] = *str2++;
   }
   if (i == size1)
   {
      str1[i-1] = '\0';
      return(0);
   }
   str1[i]  = '\0';
   return(1);
}

/***************************************************************/
int getline(register char *line, int nmax, register FILE *fp)
   /* Get at most (nmax-1) characters from file 'fp';
      put them in 'line' and append '\0';
      return the number of characters including the newline character,
         0 for an empty line, or -1 if at end-of-file.
   */
{
   register char c;
   int n;
   n = nmax;
   while ((c = getc(fp))!=EOF && --n>0)
      if ((*line++ = c) == '\0' || c == '\n') { --line; --n; break;}
   *line = '\0';
   return(nmax-n-1);
}
